# utils.py
import sys

from beartype import beartype
from loguru import logger

from feature_fabrica.exceptions import CyclicDependencyError

logger_set = False

@beartype
def setup_logger(
    log_format: str = "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | "
                      "<level>{level: <8}</level> | "
                      "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - "
                      "<level>{message}</level>",
    log_level: str = "DEBUG",
    log_file: str | None = None,
    log_rotation: str | None = None,
    log_retention: str | None = None,
    log_compression: str | None = None,
    colorize_stdout: bool = True
):
    global logger_set

    # Remove the default logger
    logger.remove()

    # Add stdout with formatting and color
    logger.add(sys.stdout, format=log_format, level=log_level, colorize=colorize_stdout)

    # Add file handler if log_file is provided
    if log_file:
        logger.add(
            log_file,
            format=log_format.replace('<level>', '').replace('</level>', ''),
            level=log_level,
            rotation=log_rotation,
            retention=log_retention,
            compression=log_compression,
        )

    logger_set = True


def get_logger(**kwargs):
    if not logger_set:
        setup_logger(**kwargs)
    return logger


def verify_dependencies(dependencies_count: dict[str, int]):
    logger = get_logger()
    if 0 in dependencies_count.values():
        loop_features = [f_name for f_name, c in dependencies_count.items() if c == 0]
        logger.debug(
            f"Cyclic dependency detected! The following features form a cycle: {loop_features}"
        )
        raise CyclicDependencyError(loop_features)



from collections.abc import Iterable

import numpy as np
from beartype import beartype

from feature_fabrica.transform.base import Transformation
from feature_fabrica.transform.utils import (
    NumericArray, NumericValue, broadcast_and_normalize_numeric_array)


class BaseReduce(Transformation):
    ufunc = None
    def __init__(self, iterable: Iterable | None = None, expects_data: bool = False, axis: int = 0):
        super().__init__()

        assert iterable or expects_data, "Either expect_data or iterable should be set!"
        self.iterable = iterable
        self.axis = axis
        if not expects_data and self.iterable:
            self.execute = self.default  # type: ignore[method-assign]
        elif expects_data and not self.iterable:
            self.execute = self.with_data  # type: ignore[method-assign]
        elif expects_data and self.iterable:
            self.execute = self.with_data_and_iterable # type: ignore[method-assign]
    @beartype
    def default(self) -> NumericArray | NumericValue:
        if self.ufunc is None:
            raise NotImplementedError()
        iterable: NumericArray = broadcast_and_normalize_numeric_array(self.iterable)
        return self.ufunc.reduce(iterable, axis=self.axis)
    @beartype
    def with_data(self, data: NumericArray | list[NumericArray]) -> NumericArray | NumericValue:
        if self.ufunc is None:
            raise NotImplementedError()
        if isinstance(data, np.ndarray):
            return self.ufunc.reduce(data, axis=self.axis)
        else:
            # Flatten the input arrays into a single contiguous array
            cells_flat = np.concatenate(data)

            # Compute the lengths and starting positions of each array
            cell_lengths = np.array([len(arr) for arr in data])
            cell_starts = np.insert(np.cumsum(cell_lengths[:-1]), 0, 0)

            # Apply the reduceat function using the starting positions
            return self.ufunc.reduceat(cells_flat, cell_starts)

    @beartype
    def with_data_and_iterable(self, data: NumericArray) -> NumericArray | NumericValue:
        if self.ufunc is None:
            raise NotImplementedError()
        data_and_iterable: NumericArray = broadcast_and_normalize_numeric_array([data] + self.iterable)
        return self.ufunc.reduce(data_and_iterable, axis=self.axis)

class SumReduce(BaseReduce):
    ufunc = np.add

class MultiplyReduce(BaseReduce):
    ufunc = np.multiply

class SubtractReduce(BaseReduce):
    ufunc = np.subtract

class DivideTransform(Transformation):
    def __init__(
        self,
        numerator: str | float | None = None,
        denominator: str | float | None = None,
    ):
        super().__init__()
        assert (
            numerator or denominator
        ), "You have to pass either numerator or denominator for computation!"

        self.numerator = numerator
        self.denominator = denominator

        if numerator and denominator:
            self.execute = self.default  # type: ignore[method-assign]
        elif numerator:
            self.execute = self.with_numerator  # type: ignore[method-assign]
        else:
            self.execute = self.with_denominator  # type: ignore[method-assign]

    @beartype
    def with_numerator(
        self, data: NumericArray | NumericValue
    ) -> NumericArray | NumericValue:
        return self.numerator / data  # type: ignore[operator]

    @beartype
    def with_denominator(
        self, data: NumericArray | NumericValue
    ) -> NumericArray | NumericValue:
        return data / self.denominator  # type: ignore[operator]

    def default(self) -> NumericArray | NumericValue:
        return self.numerator / self.denominator  # type: ignore[operator]


class ScaleFeature(Transformation):
    def __init__(self, factor: float):
        super().__init__()
        self.factor = factor

    @beartype
    def execute(self, data: NumericArray | NumericValue) -> NumericArray | NumericValue:
        return np.multiply(data, self.factor)


class LogTransform(Transformation):
    @beartype
    def execute(self, data: NumericArray | NumericValue) -> NumericArray | NumericValue:
        return np.log(data)


class ExpTransform(Transformation):
    @beartype
    def execute(self, data: NumericArray | NumericValue) -> NumericArray | NumericValue:
        return np.exp(data)


class SqrtTransform(Transformation):
    @beartype
    def execute(self, data: NumericArray | NumericValue) -> NumericArray | NumericValue:
        return np.sqrt(data)


class PowerTransform(Transformation):
    def __init__(self, power: float):
        self.power = power

    @beartype
    def execute(self, data: NumericArray | NumericValue) -> NumericArray | NumericValue:
        return data**self.power


class ZScoreTransform(Transformation):
    def __init__(self, mean: float | None = None, std_dev: float | None = None, axis: int = -1):
        super().__init__()
        self.mean = mean
        self.std_dev = std_dev
        self.axis = axis

    @beartype
    def execute(self, data: NumericArray | NumericValue) -> NumericArray | NumericValue:
        if self.mean is not None and self.std_dev is not None:
            z_normalized = (data - self.mean) / self.std_dev
        else:
            assert isinstance(data, np.ndarray), "data must be array"
            # Calculate mean of the data
            mean = np.mean(data, axis=self.axis, keepdims=True)

            # Calculate the standard deviation of the data
            std_dev = np.std(data, axis=self.axis, keepdims=True)

            # Apply Z-score normalization
            z_normalized = (data - mean) / std_dev
        return z_normalized

class ClipTransform(Transformation):
    def __init__(self, min: float, max: float):
        super().__init__()
        self.min = min
        self.max = max

    @beartype
    def execute(self, data: NumericArray | NumericValue) -> NumericArray | NumericValue:
        return np.clip(data, self.min, self.max)


class MinMaxTransform(Transformation):
    def __init__(self, min: float | None = None, max: float | None = None, axis: int = -1):
        super().__init__()
        if min is not None and max is not None:
            assert min != max
        self.min = min
        self.max = max
        self.axis = axis

    @beartype
    def execute(self, data: NumericArray | NumericValue) -> NumericArray | NumericValue:
        if self.min is not None and self.max is not None:
            min_max_normalized = (data - self.min) / (self.max - self.min)
        else:
            assert isinstance(data, np.ndarray), "data must be array"
            # Calculate min of the data
            min_ = np.min(data, axis=self.axis, keepdims=True)

            # Calculate max of the data
            max_ = np.max(data, axis=self.axis, keepdims=True)

            # Apply MinMax normalization
            min_max_normalized = (data - min_) / (max_ - min_)
        return min_max_normalized



from collections.abc import Iterable
from functools import reduce

import numpy as np
from beartype import beartype
from numpy.typing import NDArray
from sklearn.preprocessing import LabelEncoder, OneHotEncoder, OrdinalEncoder

from feature_fabrica.transform.base import Transformation
from feature_fabrica.transform.utils import StrArray, StrValue


class ToLower(Transformation):
    @beartype
    def execute(self, data: StrArray | StrValue) -> StrArray | StrValue:
        return np.char.lower(data)


class ToUpper(Transformation):
    @beartype
    def execute(self, data: StrArray | StrValue) -> StrArray | StrValue:
        return np.char.upper(data)

class ConcatenateReduce(Transformation):
    def __init__(self, iterable: Iterable | None = None, expects_data: bool = False, axis: int=-1):
        super().__init__()
        assert iterable or expects_data, "Either expect_data or iterable should be set!"
        self.iterable = iterable
        self.axis = axis
        if not expects_data and self.iterable:
            self.execute = self.default  # type: ignore[method-assign]
        elif expects_data and not self.iterable:
            self.execute = self.with_data  # type: ignore[method-assign]
        elif expects_data and self.iterable:
            self.execute = self.with_data_and_iterable # type: ignore[method-assign]

    @beartype
    def default(self) -> StrArray:
        return reduce(np.char.add, self.iterable) # type: ignore[arg-type]

    @beartype
    def with_data(self, data: StrArray | list[StrArray]) -> StrArray:
        if isinstance(data, np.ndarray):
            return np.apply_along_axis(lambda x: reduce(np.char.add, x), axis=-1, arr=data)
        else:
            return np.array([reduce(np.char.add, arr) for arr in data], dtype=str)

    @beartype
    def with_data_and_iterable(self, data: StrArray) -> StrArray:
        # TODO: make the order configurable?
        iterable_with_data = [data] + self.iterable # type: ignore[operator]
        return reduce(np.char.add, iterable_with_data)

class Strip(Transformation):
    def __init__(self, chars: str | None = None):
        super().__init__()
        self.chars = chars

    @beartype
    def execute(self, data: StrArray | StrValue) -> StrArray | StrValue:
        return np.char.strip(data, chars=self.chars)


class Split(Transformation):
    def __init__(self, delimiter: str):
        super().__init__()
        self.delimiter = delimiter

    @beartype
    def execute(self, data: StrArray | StrValue) -> np.ndarray:
        return np.char.split(data, self.delimiter)


class OneHotEncode(Transformation):
    def __init__(self, categories: list[str] | None=None, **kwargs):
        super().__init__()
        self.encoder = OneHotEncoder(dtype=np.int32, **kwargs)
        self.categories = categories

        if self.categories is not None:
            self.categories = sorted(self.categories)
            self.encoder.fit([[category] for category in self.categories])

    @beartype
    def execute(self, data: StrArray | StrValue) -> NDArray[np.int32]:
        if isinstance(data, str):
            data = np.array([data])
        # Reshape the input data to a 2D array
        data_reshaped = data.reshape(-1, 1)  # type: ignore[union-attr]
        if self.categories is not None:
            # Transform the data using the fitted encoder
            one_hot = self.encoder.transform(data_reshaped)
        else:
            one_hot = self.encoder.fit_transform(data_reshaped)

        return one_hot.toarray()


class LabelEncode(Transformation):
    def __init__(self, categories: list[str] | None = None):
        super().__init__()
        self.encoder = LabelEncoder()
        self.categories = categories
        if self.categories is not None:
            self.categories = sorted(self.categories)
            self.encoder.fit(categories)

    @beartype
    def execute(self, data: StrArray | StrValue) -> NDArray[np.int32]:
        if isinstance(data, str):
            data = np.array([data])
        if self.categories is not None:
            # Transform the data using the fitted encoder
            labels = self.encoder.transform(data)
        else:
            labels = self.encoder.fit_transform(data)

        return labels.astype(np.int32)


class OrdinalEncode(Transformation):
    def __init__(self, categories: list[str] | None = None, **kwargs):
        super().__init__()
        self.encoder = OrdinalEncoder(dtype=np.int32, **kwargs)
        self.categories = categories

        if self.categories is not None:
            self.categories = sorted(self.categories)
            self.encoder.fit([[category] for category in self.categories])

    @beartype
    def execute(self, data: StrArray | StrValue) -> NDArray[np.int32]:
        if isinstance(data, str):
            data = np.array([data])
        data_reshaped = data.reshape(-1, 1)  # type: ignore[union-attr]
        if self.categories is not None:
            ordinals = self.encoder.transform(data_reshaped)
        else:
            ordinals = self.encoder.fit_transform(data_reshaped)

        return ordinals.ravel()


class BinaryEncode(Transformation):
    def __init__(self, categories: list[str] | None = None):
        super().__init__()
        self.categories = categories
        self._cached_dict = None

        if self.categories is not None:
            self.categories = np.array(sorted(self.categories))
            self._fit(self.categories)
            cached_dict = self._build_cache()
            self.transform = np.frompyfunc(lambda x:cached_dict[x], 1, 1)

    def _build_cache(self):
        if self._cached_dict is None:
            self._cached_dict = dict(zip(self.categories, self.binary_encoded_categories))

        return self._cached_dict

    def _fit(self, data: StrArray):
        # Get the unique values and their indices in the original data
        unique_vals, inverse_indices = np.unique(data, return_inverse=True)

        # Calculate the number of binary digits needed
        num_digits = len(bin(len(unique_vals) - 1)) - 2

        # Create the binary encoded array
        binary_encoded = ((inverse_indices[:, None] & (1 << np.arange(num_digits)[::-1])) > 0).astype(np.int32)

        self.binary_encoded_categories = binary_encoded

    @beartype
    def execute(self, data: StrArray) -> NDArray[np.int32]:
        """It is useful for reducing dimensionality when dealing with categorical features that would otherwise result
        in many one-hot encoded columns."""
        assert len(data.shape) == 1, "BinaryEncode expects 1D arrays"
        if self.categories is not None:
            binary_encoded = np.stack(self.transform(data), dtype=np.int32)
        else:
            # Get the unique values and their indices in the original data
            unique_vals, inverse_indices = np.unique(data, return_inverse=True)

            # Calculate the number of binary digits needed
            num_digits = len(bin(len(unique_vals) - 1)) - 2

            # Create the binary encoded array
            binary_encoded = ((inverse_indices[:, None] & (1 << np.arange(num_digits)[::-1])) > 0).astype(np.int32)

        return binary_encoded



# core.py
import concurrent.futures
from collections import defaultdict
from collections.abc import Callable

import numpy as np
from beartype import BeartypeConf, BeartypeStrategy, beartype
from easydict import EasyDict as edict
from graphviz import Digraph
from hydra.utils import instantiate
from omegaconf import DictConfig

from feature_fabrica.models import (FeatureSpec, FeatureValue,  # noqa
                                    PromiseValue, THead, TNode)
from feature_fabrica.utils import get_logger, verify_dependencies
from feature_fabrica.yaml_parser import load_yaml

logger = get_logger()
# Dynamically create a new @slowmobeartype decorator enabling "full fat"
# O(n) type-checking.
# Type-check all items of the passed list. Do this only when you pretend
# to know in your guts that this list will *ALWAYS* be ignorably small.
slowmobeartype = beartype(conf=BeartypeConf(strategy=BeartypeStrategy.On))


class Feature:
    def __init__(self, name: str, spec: DictConfig, log_transformation_chain: bool):
        self.name = name

        self.spec = FeatureSpec(**spec)
        self.dependencies = self.spec.dependencies
        self.transformation = instantiate(self.spec.transformation)
        self.feature_value = FeatureValue(
            value=PromiseValue(), data_type=self.spec.data_type
        )

        self.log_transformation_chain = log_transformation_chain
        self.transformation_chain_head = THead()
        self.transformation_ptr = self.transformation_chain_head

        self._export_to_features = defaultdict(list) # type: ignore[var-annotated]
        if self.transformation:
            for transformation_name in self.transformation.keys():
                self._export_to_features[transformation_name]

        self.computed = False
        self._before_compute_hooks: list[Callable] = []
        self._after_compute_hooks: list[Callable] = []

    def compile(self, dependencies: dict[str, "Feature"] | None = None) -> None:
        if self.transformation:
            for (
                transformation_name,
                transformation_obj,
            ) in self.transformation.items():
                transformation_obj.compile(dependencies)

        return

    @logger.catch(reraise=True)
    def compute(self, value: np.ndarray | None = None) -> np.ndarray:
        """Compute the feature value by applying its transformation.

        Parameters
        ----------
        value : Any, optional
            The input value to the transformation.

        Returns
        -------
        np.ndarray
            The computed feature value.
        """
        # Apply the transformation function if specified
        if self.transformation:
            try:
                prev_value = value
                for (
                    transformation_name,
                    transformation_obj,
                ) in self.transformation.items():
                    if transformation_obj.expects_data:
                        result_dict = transformation_obj(prev_value)
                    else:
                        result_dict = transformation_obj()
                    prev_value = result_dict.value

                    if self._export_to_features:
                        self.export_to_features(value=result_dict.value, transformation_name=transformation_name)

                    if self.log_transformation_chain:
                        self.update_transformation_chain(
                            transformation_name, result_dict
                        )

            except Exception as e:
                if self.log_transformation_chain:
                    transformation_chain_str = self.get_transformation_chain()
                    logger.debug(transformation_chain_str)
                logger.error(
                    f"An error occurred during the transformation {transformation_name}: {e}"
                )
                raise e
            value = result_dict.value

        self.feature_value.value = value
        self.computed = True
        return self.feature_value.value  # type: ignore[attr-defined]

    @logger.catch(reraise=True)
    def __call__(self, value: np.ndarray | None = None) -> np.ndarray:
        # hooks on initial data
        for hook in self._before_compute_hooks:
            hook(data=value)
        result = self.compute(value)
        # hooks on final data
        for hook in self._after_compute_hooks:
            hook(data=result)
        return result

    def export_to_features(self, value: np.ndarray, transformation_name: str):
        feature_importers = self._export_to_features.get(transformation_name, [])
        if feature_importers:
            for feature_importer in feature_importers:
                feature_importer.pass_data(value)


    def update_transformation_chain(self, transformation_name: str, result_dict: edict):
        """Update the transformation chain with the results of the latest transformation.

        Parameters
        ----------
        transformation_name : str
            The name of the transformation.
        result_dict : edict
            The result of the transformation.
        """
        assert isinstance(
            result_dict.value, np.ndarray
        ), f"result_dict.value has to be np.ndarray, {transformation_name} might have gone wrong!"
        start_time = result_dict.start_time
        end_time = result_dict.end_time

        transformation_node = TNode(
            transformation_name=transformation_name,
            start_time=start_time,
            end_time=end_time,
        )
        transformation_node.store_hash_and_shape(result_dict.value)
        transformation_node.finilize_metrics()
        self.transformation_ptr.next = transformation_node
        self.transformation_ptr = transformation_node

    def get_transformation_chain(self) -> str:
        assert self.log_transformation_chain, f"log_transformation_chain = {self.log_transformation_chain}, turn it to True to be able to see it!"
        current = self.transformation_chain_head.next
        chain_list = []
        while current:
            chain_list.append(
                f"(Transformation: {current.transformation_name}, Hash: {current.output_hash}, Shape: {current.shape}, Time taken: {current.time_taken} seconds)"
            )
            current = current.next
        return "Transformation Chain: " + " -> ".join(chain_list)


class FeatureManager:
    def __init__(
        self,
        config_path: str,
        config_name: str,
        parallel_execution: bool = True,
        log_transformation_chain: bool = True,
    ):
        self.feature_specs: DictConfig = load_yaml(
            config_path=config_path, config_name=config_name
        )
        self.parallel_execution = parallel_execution
        self.log_transformation_chain = log_transformation_chain

        self.independent_features: list[Feature] = []
        self.dependent_features: list[Feature] = []
        self.queue: dict[int, list[Feature]] = defaultdict(list)

        self.features: edict = self._build_features()
        self.compile()

    @logger.catch(reraise=True)
    def _build_features(self) -> edict:
        """Builds features. Separates features into dependent_features and independent_features features.

        Returns
        -------
        edict
            Dictionary:
                key - > feature name (string)
                value -> feature (Feature class).
        """
        logger.info("Building features from feature definition YAML")

        features = edict()
        for name, spec in self.feature_specs.items():
            feature = Feature(
                name=name,
                spec=spec,
                log_transformation_chain=self.log_transformation_chain,
            )

            if not feature.dependencies:
                self.independent_features.append(feature)
            else:
                self.dependent_features.append(feature)

            features[name] = feature

        return features

    @logger.catch(reraise=True)
    def compile(self):
        """Identifies feature dependencies and the order in which Features are visited and computed.

        Returns
        -------
        None.
        """
        logger.info("Compiling features and feature dependencies...")

        dependencies_count = defaultdict(int)
        visited = defaultdict(int)
        # Initialize independent features
        for feature in self.independent_features:
            dependencies_count[feature.name] = 1
            visited[feature.name] = 1

        # Resolve dependent features
        for feature in self.dependent_features:
            if dependencies_count[feature.name] != 0:
                continue

            cur_feature_depends = [
                (f_name, dependencies_count[f_name]) for f_name in feature.dependencies
            ]
            if 0 not in [x[1] for x in cur_feature_depends]:
                dependencies_count[feature.name] = sum(
                    [x[1] for x in cur_feature_depends]
                )
            else:
                # Handle unresolved dependencies using a stack
                stack = [
                    f_name
                    for f_name in feature.dependencies
                    if dependencies_count[f_name] == 0
                ]
                while stack:
                    f_node_name = stack.pop()
                    if visited[f_node_name]:
                        continue

                    # Mark this node as visited
                    visited[f_node_name] = 1

                    # Get the feature object by its name
                    f_node = self.features[f_node_name]

                    # Resolve dependencies of this node
                    node_feature_depends = [
                        (f_name, dependencies_count[f_name])
                        for f_name in f_node.dependencies
                    ]

                    if 0 in [x[1] for x in node_feature_depends]:
                        # If there are still unresolved dependencies, push back on stack
                        stack.append(f_node_name)
                        for dep_name, dep_count in node_feature_depends:
                            if dep_count == 0 and not visited[dep_name]:
                                stack.append(dep_name)
                    else:
                        # All dependencies resolved, update count
                        dependencies_count[f_node_name] = sum(
                            [x[1] for x in node_feature_depends]
                        )

                # Finally, update the current feature's dependency count
                dependencies_count[feature.name] = sum(
                    [dependencies_count[f_name] for f_name in feature.dependencies]
                )

        verify_dependencies(dependencies_count)
        for f_name, level in dependencies_count.items():
            self.queue[level].append(self.features[f_name])

        self.compile_features()

    def compile_features(self):
        def compile_feature(feature: Feature):
            if not feature.dependencies:
                # Independent feature
                feature.compile(None)
            else:
                # Dependent feature
                dependencies = {
                    f_name: self.features[f_name] for f_name in feature.dependencies
                }
                feature.compile(dependencies=dependencies)

        for priority in sorted(self.queue.keys()):
            cur_features = self.queue[priority]
            if self.parallel_execution:
                with concurrent.futures.ThreadPoolExecutor() as executor:
                    future_to_feature = {
                        executor.submit(compile_feature, feature): feature
                        for feature in cur_features
                    }
                    for future in concurrent.futures.as_completed(future_to_feature):
                        future.result()
            else:
                for feature in cur_features:
                    compile_feature(feature=feature)

    def compute_single_feature(self, feature: Feature, value: np.ndarray | None = None):
        if value is not None:
            result = feature(value=value)
        else:
            result = feature()
        return feature.name, result

    @slowmobeartype
    def compute_features_with_validation(
        self, data_keys: list[str], data_values: list[np.ndarray]
    ) -> edict:
        """

        Parameters
        ----------
        data : dict[str, Any]
            Data point.

        Returns
        -------
        Dictionary
            Processed data point with derived features as well.

        """
        data = dict(zip(data_keys, data_values))
        results = {}

        for priority in sorted(self.queue.keys()):
            cur_features = self.queue[priority]

            if self.parallel_execution:
                with concurrent.futures.ThreadPoolExecutor() as executor:
                    future_to_feature = {
                        executor.submit(
                            self.compute_single_feature,
                            feature,
                            data[feature.name] if feature.name in data else None,
                        ): feature
                        for feature in cur_features
                    }
                    for future in concurrent.futures.as_completed(future_to_feature):
                        feature_name, result = future.result()
                        results[feature_name] = result
            else:
                for feature in cur_features:
                    feature_name, result = self.compute_single_feature(
                        feature=feature,
                        value=data[feature.name] if feature.name in data else None,
                    )
                    results[feature_name] = result

        return edict(results)

    def compute_features(self, data: dict[str, np.ndarray]) -> edict:
        return self.compute_features_with_validation(
            list(data.keys()), list(data.values())
        )

    def get_visual_dependency_graph(
        self, save_plot: bool = False, output_file: str = "feature_dependencies"
    ):
        dot = Digraph(comment="Feature Dependencies")

        # Add nodes
        for feature in self.features.values():
            dot.node(feature.name)

        # Add edges
        for feature in self.features.values():
            for dependency in feature.dependencies:
                dot.edge(dependency, feature.name)

        if save_plot:
            # Save and render the graph
            dot.render(output_file, format="png")
            logger.info(f"Dependencies graph saved as {output_file}.png")
        return dot



from beartype import beartype

from feature_fabrica.core import Feature
from feature_fabrica.models import FeatureValue
from feature_fabrica.transform.base import Transformation
from feature_fabrica.transform.utils import StrOrNumArray


class FeatureImporter(Transformation):
    @beartype
    def __init__(self, feature: str, transform_stage: str | None = None):
        """
        Parameters
        ----------
        feature : str
            The name of the feature to import.
        transform_stage : str, optional
            The specific stage of transformation from which to import data. If not provided (None),
            the initial raw feature value before any transformations will be used.
        """
        super().__init__()
        self.feature = feature
        self.transform_stage = transform_stage
        self.data = None

    @beartype
    def compile(self, features: dict[str, Feature]) -> bool:
        if self.feature not in features:
            raise ValueError(f"Feature '{self.feature}' not found.")

        feature_obj = features[self.feature]
        if self.transform_stage is not None:
            if self.transform_stage not in feature_obj.transformation:
                raise ValueError(f"Transform stage '{self.transform_stage}' not found in feature '{self.feature}'.")
            feature_obj._export_to_features[self.transform_stage].append(self)
        else:
            feature_obj._before_compute_hooks.append(self.pass_data)
        return False

    @beartype
    def pass_data(self, data: StrOrNumArray) -> None:
        self.data = data

    @beartype
    def execute(self) -> StrOrNumArray:
        if self.data is None:
            raise ValueError("Data has not been set.")
        return self.data.value if isinstance(self.data, FeatureValue) else self.data



# models.py
import hashlib
import re
from typing import Any, Optional

import numpy as np
from pydantic import BaseModel, ConfigDict, Field, root_validator, validator


class PromiseValue:
    value: np.ndarray | None = None


class FeatureSpec(BaseModel):
    description: str
    data_type: str
    dependencies: list[str] | None = Field(default_factory=list)
    transformation: dict | None = Field(default_factory=dict)

    @validator("data_type")
    def validate_data_type(cls, v):
        try:
            # Check if the data_type is a valid type
            if not (getattr(np, v, None)):
                raise ValueError(
                    f"Invalid data_type specified: {v}, it should be in numpy"
                )
        except Exception as e:
            raise ValueError(f"Invalid data_type: {v}, error: {str(e)}")
        return v


class FeatureValue(np.lib.mixins.NDArrayOperatorsMixin, BaseModel, validate_assignment=True):  # type: ignore[call-arg]
    model_config = ConfigDict(arbitrary_types_allowed=True)

    value: np.ndarray | PromiseValue = Field(default=None)
    data_type: str

    @root_validator(pre=True)
    def validate_value(cls, values):
        v = values.get("value")
        data_type = values.get("data_type")
        # allow PromiseValue only
        if isinstance(v, PromiseValue):
            return values
        # Get the expected data type from builtins or numpy
        expected_type = getattr(np, data_type, None)

        if expected_type is None:
            raise ValueError(f"Unsupported data type '{data_type}', use numpy typing!")

        # Validate that the array has the correct data type
        actual_type_name = v.dtype.name
        actual_type, actual_precision = re.findall(r'\D+|\d+', actual_type_name)

        expected_type_list = re.findall(r'\D+|\d+', data_type)

        if actual_type not in data_type:
            raise ValueError(
                f"Array dtype '{v.dtype}' does not match expected type '{data_type}'"
            )
        elif len(expected_type_list) > 1 and actual_precision != expected_type_list[1]:
            # convert to desired type if compatible
            v = v.astype(expected_type)

        # Update the value in the values dictionary
        values["value"] = v
        return values

    def __array__(self, dtype=None, copy=None):
        # Automatically converts to np.ndarray when passed to a function that expects an array
        if dtype:
            return self.value.astype(dtype)
        return self.value.copy() if copy else self.value

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        # Convert inputs to their underlying values if they are FeatureValue instances
        inputs = tuple(x.value if isinstance(x, FeatureValue) else x for x in inputs)

        # Perform the operation using the ufunc
        result = getattr(ufunc, method)(*inputs, **kwargs)
        return result

    def __array_function__(self, func, types, args, kwargs):
        # Convert args to their underlying values if they are FeatureValue instances
        args = tuple(x.value if isinstance(x, FeatureValue) else x for x in args)
        # Perform the operation using the function
        result = func(*args, **kwargs)
        return result

    def __getitem__(self, idx):
        return self.value[idx]

    def __getattr__(self, name):
        return getattr(self.value, name)

    def __repr__(self):
        return f"FeatureValue(value={self.value})"

class TNode(BaseModel):
    transformation_name: str
    start_time: float
    end_time: float
    shape: tuple | None = None
    time_taken: float | None = None
    output_hash: str | None = None
    next: Optional["TNode"] = None  # Forward reference

    def to_dict(self) -> dict[str, Any]:
        # Convert the node and its next nodes to a dictionary
        node_dict = {
            "value": self.value,
            "transformation_name": self.transformation_name,
        }
        if self.next:
            node_dict["next"] = self.next.to_dict()
        else:
            node_dict["next"] = None
        return node_dict

    def compute_hash(self, data: np.ndarray) -> str:
        data_bytes = data.tobytes()
        return hashlib.sha256(data_bytes).hexdigest()

    def store_hash_and_shape(self, output_data: np.ndarray):
        self.shape = output_data.shape
        self.output_hash = self.compute_hash(output_data)

    def finilize_metrics(self) -> None:
        self.time_taken = self.end_time - self.start_time


class THead(BaseModel):
    next: TNode | None = None



import numpy as np
from beartype import beartype
from scipy import stats

from feature_fabrica.transform.base import Transformation
from feature_fabrica.transform.utils import NumericArray, StrOrNumArray


@beartype
def mean_reduction(data: NumericArray | list[NumericArray], axis: int) -> NumericArray:
    if isinstance(data, np.ndarray):
        return np.mean(data, axis=axis)
    else:
        # Flatten the input arrays into a single contiguous array
        cells_flat = np.concatenate(data)

        # Compute the lengths and starting positions of each array
        cell_lengths = np.array([len(arr) for arr in data])
        cell_starts = np.insert(np.cumsum(cell_lengths[:-1]), 0, 0)
        return np.add.reduceat(cells_flat, cell_starts) / cell_lengths

@beartype
def min_reduction(data: NumericArray | list[NumericArray], axis: int) -> NumericArray:
    if isinstance(data, np.ndarray):
        return np.min(data, axis=axis)
    else:
        return np.array([np.min(arr) for arr in data], dtype=float)

@beartype
def max_reduction(data: NumericArray | list[NumericArray], axis: int) -> NumericArray:
    if isinstance(data, np.ndarray):
        return np.max(data, axis=axis)
    else:
        return np.array([np.max(arr) for arr in data], dtype=float)

@beartype
def median_reduction(data: NumericArray | list[NumericArray], axis: int) -> NumericArray:
    if isinstance(data, np.ndarray):
        return np.median(data, axis=axis)
    else:
        return np.array([np.median(arr) for arr in data], dtype=float)

@beartype
def mode_reduction(data: NumericArray | list[NumericArray], axis: int) -> NumericArray:
    if isinstance(data, np.ndarray):
        # use LabelEncode on your feature first: https://github.com/scipy/scipy/issues/15551
        mode_result = stats.mode(data, axis=axis)
        # Extract the mode from the result
        return mode_result.mode
    else:
        return np.array([stats.mode(arr, axis=-1).mode for arr in data], dtype=float)

DEAFULT_REDUCTIONS = dict(
        mean = mean_reduction,
        min = min_reduction,
        max = max_reduction,
        median = median_reduction,
        mode = mode_reduction
    )

class GroupByReduce(Transformation):
    @beartype
    def __init__(self, key_feature: str, reduce_func: Transformation | str = "mean", axis: int = -1):
        super().__init__()
        if isinstance(reduce_func, str):
            assert reduce_func in ["mean", "mode", "min", "max", "median"]
            reduce_func = DEAFULT_REDUCTIONS[reduce_func]
        self.key_feature = key_feature
        self.reduce_func = reduce_func
        self.axis = axis

    @beartype
    def execute(self, data: StrOrNumArray) -> StrOrNumArray:
        """Groups data array by key_feature values and applies a reduction transformation.

        Assigns reduced values to the original positions to preserve the original shape.
        """
        if isinstance(self.reduce_func, Transformation):
            assert self.reduce_func.expects_data
        #assert isinstance(self.key_feature, FeatureValue), "key_feature must be an existing feature!"
        assert self.key_feature.shape == data.shape # type: ignore[attr-defined]
        # FeatureValue has weird behavior with np.unique
        key_feature_value = self.key_feature.value # type: ignore[attr-defined]

        sorted_idxs = key_feature_value.argsort() # type: ignore[attr-defined]
        data_sorted = data[sorted_idxs]
        key_feature_sorted = key_feature_value[sorted_idxs]

        _, unqiue_indeces, counts = np.unique(key_feature_sorted, return_index=True, return_counts=True)
        data_aggregated =  np.split(data_sorted, unqiue_indeces[1:])

        if np.all(counts == counts[0]):
            data_aggregated = np.array(data_aggregated, dtype=data.dtype)

        if isinstance(self.reduce_func, Transformation):
            data_reduced = self.reduce_func.execute(data_aggregated)
        else:
            data_reduced = self.reduce_func(data_aggregated, self.axis) # type: ignore[operator]
        # Use np.unique to find the unique key values and the counts for each
        _, inverse_indices = np.unique(key_feature_value, return_inverse=True)
        # Use np.repeat to assign the reduced values to the original positions
        result = data_reduced[inverse_indices]

        return result
